@@ -17,13 +17,13 @@ item *do_item_alloc_pull(const size_t ntotal, const unsigned int id);
 void item_free(item *it);
 bool item_size_ok(const size_t nkey, const client_flags_t flags, const int nbytes);
 
-int  do_item_link(item *it, const uint32_t hv);     /** may fail if transgresses limits */
+int  do_item_link(item *it, const uint32_t hv, const uint64_t cas);     /** may fail if transgresses limits */
 void do_item_unlink(item *it, const uint32_t hv);
 void do_item_unlink_nolock(item *it, const uint32_t hv);
 void do_item_remove(item *it);
 void do_item_update(item *it);   /** update LRU time to current and reposition */
 void do_item_update_nolock(item *it);
-int  do_item_replace(item *it, item *new_it, const uint32_t hv);
+int  do_item_replace(item *it, item *new_it, const uint32_t hv, const uint64_t cas);
 void do_item_link_fixup(item *it);
 
 int item_is_flushed(item *it);
@@ -47,6 +47,7 @@ struct lru_pull_tail_return {
 int lru_pull_tail(const int orig_id, const int cur_lru,
         const uint64_t total_bytes, const uint8_t flags, const rel_time_t max_age,
         struct lru_pull_tail_return *ret_it);
+void item_flush_expired(void);
 
 /*@null@*/
 char *item_cachedump(const unsigned int slabs_clsid, const unsigned int limit, unsigned int *bytes);
@@ -57,8 +58,6 @@ void item_stats_totals(ADD_STAT add_stats, void *c);
 /*@null@*/
 void item_stats_sizes(ADD_STAT add_stats, void *c);
 void item_stats_sizes_init(void);
-void item_stats_sizes_enable(ADD_STAT add_stats, void *c);
-void item_stats_sizes_disable(ADD_STAT add_stats, void *c);
 void item_stats_sizes_add(item *it);
 void item_stats_sizes_remove(item *it);
 bool item_stats_sizes_status(void);
